char infomessage[] = "\
\n\
                   MPEG Video Software Decoder\n\
              (Motif-Version 2.0.1; March 15, 1994)\n\
\n\
        Lawrence A. Rowe, Ketan Patel, and Brian Smith\n\
 Computer Science Division-EECS, Univ. of Calif. at Berkeley\n\
\n\
                  Motif Interface: Daeron Meyer\n\
           The Geometry Center, University of Minnesota\n\
\n\
          Thanks to Mike Peter Bretz and Thomas Woerner,\n\
                University of Tuebingen, Germany,\n\
         for adding the \"Save\" button and stdin handling.\n\
\n\
There are several mailing lists established for messages about\n\
the decoder:\n\
\n\
mpeg-list-dist@CS.Berkeley.EDU\n\
   General information on the decoder for everyone interested\n\
   should be sent to this list.  This should become active after\n\
   11/20/92\n\
\n\
mpeg-list-request@CS.Berkeley.EDU\n\
   Requests to join or leave the list should be sent to this\n\
   address. The subject line should contain the single word\n\
   ADD or DELETE.\n\
\n\
mpeg-bugs@CS.Berkeley.EDU\n\
   Problems, questions, or patches should be sent to this address.\n\
\n\
If you have any questions or bug reports for the motif version\n\
(and I'm sure there will be plenty) please send mail to:\n\
\n\
   daeron@geom.umn.edu\n\
\n\
Also check out the online web documentation for the motif version:\n\
\n\
   http://www.geom.umn.edu/docs/mpeg_play/mpeg_play.html\n\
";
